/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafkarest.entities;

import com.google.auto.value.AutoValue;

@AutoValue
public abstract class PartitionReplica {

  PartitionReplica() {
  }

  public abstract String getClusterId();

  public abstract String getTopicName();

  public abstract int getPartitionId();

  public abstract int getBrokerId();

  public abstract boolean isLeader();

  public abstract boolean isInSync();

  public static PartitionReplica create(
      String clusterId,
      String topicName,
      int partitionId,
      int brokerId,
      boolean isLeader,
      boolean isInSync) {
    return new AutoValue_PartitionReplica(
        clusterId, topicName, partitionId, brokerId, isLeader, isInSync);
  }
}
